#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            oneWire.py
#
# Licence:         Donationware, distributed as Freeware
#
#                  If you find this software useful to you and you would
#                  like to help the developer to maintain it also in the future, 
#                  please donate a small amount of your liking to the 
#                  PayPal account: olammi@iki.fi.
#
#                  This donation is not required to use the software and the 
#                  software is licenced as freeware.
#
#                  To individual and non-commercial use only! Script
#                  and source code may be used freely for non-commercial
#                  purposes. Any commercial usage requires permission or
#                  licence from the author. 
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.0b
#
# Date:            06.04.2009
#
# Description:     Library to communicate with the 1-wire devices using OWFS
#                  and DigiTemp
#                  
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.3)
#
# Version history: ** 24.02.2009 v0.9a (Olli Lammi) **
#                  First beta version for testers. 
#
#                  ** 25.02.2009 v0.9c (Olli Lammi) **
#                  Logging and import changes 
#
#                  ** 06.04.2009 v1.0a (Olli Lammi) **
#                  Added DigiTemp support. 
#
#                  ** 06.04.2009 v1.0b (Olli Lammi) **
#                  Corrected DigiTemp support. 
#
###########################################################################

# Imports

import sys, os 
import string, subprocess

import log

###########################################################################

# Constants

OWTYPE_FLOAT = 1
OWTYPE_INT = 2


###########################################################################

# Classes

class OneWireFS(log.Logging):
    def __init__(self):
        log.Logging.__init__(self, 'OneWireFS')
        self.OWFS_DEVICES = {}

    def setDevices(self, devmap):
        self.OWFS_DEVICES = devmap

    def runQueryCommand(self, cmd):
        try:
            item = self.OWFS_DEVICES[cmd]
        except:
            self.Log("ERROR: Invalid one wire key: " + cmd)
            return ""

        type = item[0]
        fname = item[1]

        if len(fname) <= 0:
            self.Log("ERROR: Empty sensor file for key: " + cmd)
            return ""

        self.Debug("Running oneWire query from OWFS file " + fname)
        data = ""
        try:
            infile = open(fname, 'r')
            data = infile.read()
            infile.close()
        except:
            self.Log("ERROR: Error reading sensor file: " + fname)
            return ""
        self.Debug("Received oneWire data: " + repr(data))

        if type == OWTYPE_FLOAT:
            data = string.strip(data)
            if len(data) > 0:
                data = "%.2f" % float(data)
        elif type == OWTYPE_INT:
            data = string.strip(data)
            if len(data) > 0:
                data = "%d" % int(data)
 
        return data


class OneWireDigiTemp(log.Logging):
    def __init__(self, dtbin, dtconf, dtport):
        log.Logging.__init__(self, 'OneWireDigiTemp')
        self.DT_BINARY = dtbin
        self.DT_RC = dtconf
        self.DT_PORT = dtport

    def runQueryCommand(self, cmd):
        digitempcmd = self.DT_BINARY + ' -q -c ' + self.DT_RC + ' -s ' + self.DT_PORT + ' -t ' + cmd
        digitempcmd = digitempcmd + ' -o "%.2C" -O "%C" -H "%h"'
        self.Debug("Reading data from DigiTemp using command: " + digitempcmd)
        stat = -1
        output = ['', '']
        try:      
            sb = subprocess.Popen(digitempcmd, shell=True, stdout=subprocess.PIPE, stderr=subprocess.PIPE)
            output = sb.communicate()
            stat = sb.returncode
        except:
            stat = -1
        if stat != 0:
            temps = "ERROR: Error reading from DigiTemp\n"
            temps = temps + "  DigiTemp command: " + digitempcmd + "\n"
            temps = temps + "  DigiTemp output: " + output[0] + output[1]
            self.Log(temps)                
            data = ""
        else:
            data = string.strip(output[0])
        self.Debug("Received oneWire data: " + repr(data))

        if len(data) > 0:
            rivit = string.split(data, '\n')
            data = ''
            for rivi in rivit:
                rivi = string.strip(rivi)                
                if len(rivi) > 0:
                    try:    
                        data = "%.2f" % float(rivi)
                        return data
                    except:
                        continue
                    
        return data

