#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            oumanSerial.py
#
# Licence:         Donationware, distributed as Freeware
#
#                  If you find this software useful to you and you would
#                  like to help the developer to maintain it also in the future, 
#                  please donate a small amount of your liking to the 
#                  PayPal account: olammi@iki.fi.
#
#                  This donation is not required to use the software and the 
#                  software is licenced as freeware.
#
#                  To individual and non-commercial use only! Script
#                  and source code may be used freely for non-commercial
#                  purposes. Any commercial usage requires permission or
#                  licence from the author. 
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         0.9f
#
# Date:            30.04.2010
#
# Description:     Library to communicate with the Ouman controller using the
#                  Ouman native serial communication protocol. Mainly used to 
#                  read measurement data from the Ouman controller. Will contain
#                  also methods to read and acknowledge alarms and read and alter
#                  controller configuration variables.
#                  
# Supported devices: 
#                  At the moment the supported Ouman EH-200 series controller
#                  devices are:
#
#                     EH202S
#                     EH202G
#                     EH203G
#                     EH203T
#                     EH201T  
#                     EH201L 
#                     EH201V
#                     EH201G
#                   
#                  If your equipment is not listed, you might still be able to 
#                  use one of the models above to do some logging. You can
#                  also contact the author to add support to a new device or
#                  to improve support for an existing one. 
#
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.3)
# 
#                  Python Serial Port Extension: pySerial  
#                  (http://pyserial.wiki.sourceforge.net/pySerial)
#
# Version history: ** 24.02.2009 v0.9a (Olli Lammi) **
#                  First beta version for testers. 
#
#                  ** 25.02.2009 v0.9c (Olli Lammi) **
#                  Logging and import changes 
#
#                  ** 02.03.2009 v0.9d (Olli Lammi) **
#                  Added locking using a lock file. Changed the module to
#                  open the port only when accessing the serial module.
#
#                  ** 02.06.2009 v0.9e (Olli Lammi) **
#                  Added profile for EH202G (Geopro GS)
#
#                  ** 30.04.2010 v0.9f (Olli Lammi) **
#                  Removed handling of twos complement to 
#                  external module binary.py.
#
###########################################################################

# Imports

import sys, os 
import string
import tempfile

import log
import lockfile
import binary

# pySerial module, not loaded until init of OumanSerial class instance
serial = None

###########################################################################

# Constants

BAUDRATE = 4800 
TIMEOUT = 2

LOCKFILENAME = tempfile.gettempdir() + os.sep + 'lock_oumanSerial_'

###########################################################################

# Device configuration data

TYPE_TEMP100 = 1 
TYPE_INT16 = 2

OUMAN_DEVICES = { 'EH202S': [ ['Ulkolampotila', 18, TYPE_TEMP100], \
                              ['L1 huone', 21, TYPE_TEMP100], \
                              ['L1 menovesi', 20, TYPE_TEMP100], \
                              ['L2 menovesi', 26, TYPE_TEMP100], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100], \
                              ['Varaaja ylatulo', 24, TYPE_TEMP100], \
                              ['Varaaja yla', 33, TYPE_TEMP100], \
                              ['Varaaja keski', 23, TYPE_TEMP100], \
                              ['Varaaja ala', 34, TYPE_TEMP100], \
                              ['Maaliuos', 25, TYPE_TEMP100], \
                              ['Tilamuuttuja', 45, TYPE_INT16] \
                            ], \
                  'EH202G': [ ['Ulkolampotila', 18, TYPE_TEMP100], \
                              ['L1 huone', 21, TYPE_TEMP100], \
                              ['L1 menovesi', 20, TYPE_TEMP100], \
                              ['L2 menovesi', 26, TYPE_TEMP100], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100], \
                              ['Varaaja ala', 24, TYPE_TEMP100], \
                              ['Varaaja yla', 33, TYPE_TEMP100], \
                              ['L1 paluuvesi', 23, TYPE_TEMP100], \
                              ['Lammitys varaaja ala', 34, TYPE_TEMP100], \
                              ['Maaliuos', 25, TYPE_TEMP100], \
                              ['Tilamuuttuja', 45, TYPE_INT16] \
                            ], \
                  'EH203G': [ ['Ulkolampotila', 18, TYPE_TEMP100], \
                              ['L1 huone', 21, TYPE_TEMP100], \
                              ['L1 menovesi', 20, TYPE_TEMP100], \
                              ['L2 menovesi', 26, TYPE_TEMP100], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100], \
                              ['Varaaja yla', 33, TYPE_TEMP100], \
                              ['Varaaja ala', 34, TYPE_TEMP100], \
                              ['Maaliuos', 25, TYPE_TEMP100], \
                              ['LV menovesi', 24, TYPE_TEMP100], \
                              ['Kuumakaasu', 41, TYPE_TEMP100], \
                              ['Tilamuuttuja', 45, TYPE_INT16] \
                            ], \
                  'EH203T': [ ['Ulkolampotila', 18, TYPE_TEMP100], \
                              ['L1 huone', 21, TYPE_TEMP100], \
                              ['L1 menovesi', 20, TYPE_TEMP100], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100], \
                              ['Varaaja yla', 33, TYPE_TEMP100], \
                              ['Varaaja ala', 34, TYPE_TEMP100], \
                              ['Maaliuos', 25, TYPE_TEMP100], \
                              ['LV menovesi', 24, TYPE_TEMP100], \
                              ['Kuumakaasu', 41, TYPE_TEMP100], \
                              ['Tilamuuttuja', 45, TYPE_INT16] \
                            ], \
                  'EH201T': [ ['Ulkolampotila', 18, TYPE_TEMP100], \
                              ['L1 huone', 21, TYPE_TEMP100], \
                              ['L1 menovesi', 20, TYPE_TEMP100], \
                              ['Varaaja yla', 33, TYPE_TEMP100], \
                              ['Varaaja ala', 34, TYPE_TEMP100], \
                              ['Kuumakaasu', 41, TYPE_TEMP100], \
                              ['Tilamuuttuja', 45, TYPE_INT16] \
                            ], \
                  'EH201L': [ ['Ulkolampotila', 18, TYPE_TEMP100], \
                              ['L1 huone', 21, TYPE_TEMP100], \
                              ['L1 menovesi', 20, TYPE_TEMP100], \
                              ['L1 paluuvesi', 23, TYPE_TEMP100], \
                              ['Mittaus 9', 33, TYPE_TEMP100], \
                              ['Mittaus 10', 34, TYPE_TEMP100], \
                              ['Mittaus 11', 41, TYPE_TEMP100], \
                              ['Tilamuuttuja', 45, TYPE_INT16] \
                            ], \
                  'EH201V': [ ['LV menovesi', 24, TYPE_TEMP100], \
                              ['LV kiertovesi', 25, TYPE_TEMP100], \
                              ['Mittaus 9', 33, TYPE_TEMP100], \
                              ['Mittaus 10', 34, TYPE_TEMP100], \
                              ['Mittaus 11', 41, TYPE_TEMP100], \
                              ['Tilamuuttuja', 45, TYPE_INT16] \
                            ], \
                  'EH201G': [ ['Ulkolampotila', 18, TYPE_TEMP100], \
                              ['L1 huone', 21, TYPE_TEMP100], \
                              ['L1 menovesi', 20, TYPE_TEMP100], \
                              ['Vapaa mittaus', 27, TYPE_TEMP100], \
                              ['Varaaja yla', 33, TYPE_TEMP100], \
                              ['Varaaja ala', 34, TYPE_TEMP100], \
                              ['Kuumakaasu', 41, TYPE_TEMP100], \
                              ['Tilamuuttuja', 45, TYPE_INT16] \
                            ] \
                }

TILA_MASK_DIGI1 = 1 
TILA_MASK_DIGI2 = 2 
TILA_MASK_DIGI3 = 4 
TILA_MASK_SAHKOVASTUS = 8 
TILA_MASK_KOMPRESSORI = 16


###########################################################################

# Classes

class OumanSerial(log.Logging):
    def __init__(self, port, device):
        log.Logging.__init__(self, 'OumanSerial')
        
        global serial
        serial = __import__('serial') 
        
        self.SERPORT = port
        self.OUMAN_DEVICE = device
        self.serio = 0
          
        self.LOCK = lockfile.LockFile(LOCKFILENAME + string.replace(self.SERPORT, os.sep, '_'))
    
    def openPort(self):
        if not OUMAN_DEVICES.has_key(self.OUMAN_DEVICE):
            self.Log("ERROR: Invalid OUMAN device type: " + self.OUMAN_DEVICE)
            return 0

        if not self.LOCK.lock():
            self.Log("ERROR: Unable to aquire lockfile for oumanSerial port: " + self.LOCK.getName())
            return 0

        try:
            self.serio = serial.Serial(port=self.SERPORT,baudrate=BAUDRATE, \
                                bytesize=serial.EIGHTBITS, \
                                parity=serial.PARITY_NONE, stopbits=serial.STOPBITS_ONE, \
                                xonxoff=0, rtscts=0, dsrdtr=0, timeout=TIMEOUT)
            if not self.serio.isOpen():
                self.serio = 0
                self.LOCK.free()
                return 0 
        except:
            self.serio = 0
            self.LOCK.free()
            return 0
        
        self.Debug("Opened oumanSerial port: " + self.SERPORT)
        return 1
        
    def closePort(self):
        if self.serio:
            try:
                self.serio.close()
            except:
                pass
            self.serio = 0
        self.LOCK.free()
        self.Debug("Closed oumanSerial port: " + self.SERPORT)

    def isOpen(self):
        if self.serio and self.serio.isOpen():
            return 1
        return 0

    def runQueryCommand(self, cmd):
        id = 0
        type = 0
        for t in OUMAN_DEVICES[self.OUMAN_DEVICE]:
            if t[0] == cmd:
                id = t[1]
                type = t[2]
                break

        if type == 0:
            self.Log("ERROR: Invalid OUMAN key: " + cmd)
            return ""

        res = self.runQueryId(id)
        if len(res) > 0:
            return convertOumanMessage(type, res)
        return ""

    def runQueryId(self, id):
        self.Debug("Running oumanSerial query with id " + repr(id))
        if not self.isOpen():
            if not self.openPort():
                self.Log("ERROR: Cannot open serial port.")
                return ""
            (stat, res) = serialQuery(self.serio, id)
            self.closePort()
        else:
            (stat, res) = serialQuery(self.serio, id)
        self.Debug("Received ouman response: " + repr(res))
        
        if stat == 1:
            return res
        elif stat == -1:
            self.Log("ERROR: Timeout reading message from Ouman")
            return ""
        elif stat == -2:
            self.Log("ERROR: Error in return message from Ouman")
            return ""


###########################################################################

# Functions

# returns (status, response)
# status: 1 = ok, -1 = timeout, -2 = error in response
def serialQuery(serio, id):
    crc = (0x81 + 0x02 + 0x00 + id) & 0xFF
    sendbuff = '\x02\x81\x02\x00' + chr(id) + chr(crc)

    serio.write(sendbuff)
    serio.flush()

    res = ""
    temp = " "
    while len(temp) > 0:
        temp = serio.read()
        if len(temp) <= 0:
            if checkOumanMessage(res) == 1:
                return (1, res)
            else:
                return (-1, '')
        res = res + temp
        stat = checkOumanMessage(res)
        if stat == 1:
            return (1, res)
        elif stat == 2:
            return (-2, '')

# returns int
# 0 = ok, but not ready
# 1 = ok
# 2 = error
def checkOumanMessage(data):
    if len(data) <= 0:
        return 0

    # first byte is 0x02
    if data[0] != chr(0x02):
        return 2

    # ignore second byte
    if len(data) < 3:
        return 0

    # byte 3 is content lenght
    dlen = ord(data[2])

    # if package is already longer than it should
    if len(data) > dlen + 4:
        return 2

    if len(data) < dlen + 4:
        return 0

    # if package crc does not match
    crc = 0
    for i in range(1, len(data)-1):
        crc = crc + ord(data[i])
    if (crc & 0xFF) != ord(data[len(data)-1]):
        return 2

    return 1


def getOumanDataPart(data):
    if checkOumanMessage(data) != 1:
        return ""

    dlen = ord(data[2])
    return data[3: 3+dlen]


def convertOumanMessage(type, data):
    # get data part and remove id part
    data = getOumanDataPart(data)

    if len(data) <= 2:
        return ""
    data = data[2:]

    if type == TYPE_TEMP100:
        temp = 0
        temp = (ord(data[0]) << 8) + ord(data[1])
        temp = binary.twosComplementToInt(temp, 16)
        temp = "%.2f" % (temp / 100.0)
        return temp
    elif type == TYPE_INT16:
        temp = (ord(data[0]) << 8) + ord(data[1])
        temp = "%d" % temp
        return temp

    return ""
