#!env python
# -*- coding: iso-8859-1 -*-
###########################################################################
# 
# File:            taloLogger.py
#
#
# Licence:         Donationware, distributed as Freeware
#
#                  If you find this software useful to you and you would
#                  like to help the developer to maintain it also in the future, 
#                  please donate a small amount of your liking to the 
#                  PayPal account: olammi@iki.fi.
#
#                  This donation is not required to use the software and the 
#                  software is licenced as freeware.
#
#                  To individual and non-commercial use only! Script
#                  and source code may be used freely for non-commercial
#                  purposes. Any commercial usage requires permission or
#                  licence from the author. 
#
#
HELP_TEXT = """
# Usage:           python taloLogger.py --help
#                  python taloLogger.py [-d] [-v] [-l] [-f configuration_file]
#
#                  Options (options override the configuration file):
#                    --help    - Prints out a help and exits.
#
#                    -d        - Runs the application in daemon mode. Command will
#                                return to shell and a daemon will be forked to
#                                run in the background. 
#
#                    -v        - Enable debug/verbose logging.
#
#                    -l        - Console log mode. 
#
#                    -f configuration_file - Use the given filename as configuration
#                                file for the application.
##
#                  Parameters:
#                                None.
#
#                  Configuration:
#                                See file taloLogger.conf for example configuration
#                                and configuration keys.
#
"""
#
# Author:          Olli Lammi (olammi@iki.fi)
#
# Version:         1.3a
#
# Date:            16.11.2010
#
# Functions:       -
#                   
# Description:     House automation logger application for Python. The primary
#                  driver has been to produce a logger application to read Ouman
#                  control device and 1-wire sensor values and log them 
#                  periodically to MySql database. 
#                  
#                  Can be extended to log to different data storages: file,
#                  rrdtool, database, etc. Suggest a new data storage for the 
#                  author if you need a specific module.
#
#                  Currently is able to read Ouman EH-200 series controllers 
#                  using the native Ouman serial protocol and 1-wire devices
#                  through the OWFS interface or Digitemp utilities.
#
# Requirements:    Python interpreter 2.4 or newer (www.python.org)
#                  (tested with 2.4.4)
# 
#                  For Ouman and Thermia controller logging:
#                    Python Serial Port Extension: pySerial  
#                    (http://pyserial.wiki.sourceforge.net/pySerial)
#
#                  For MySQL database data storage:
#                    MySQL-python library (MySQLdb) 
#                    (http://sourceforge.net/projects/mysql-python)
#
#                  For 1-wire data sources one of below:
#                    OWFS (http://www.owfs.org/) 
#                    DigiTemp (http://www.digitemp.com/)
#
#                  For RRDtool data storage:
#                    RRDtool (http://oss.oetiker.ch/rrdtool/)
#
#                  For SQLite database data storage:
#                    Python interpreter 2.5 or newer with builtin
#                    SQLite3 support.
#
# Version history: ** 24.02.2009 v0.9a (Olli Lammi) **
#                  First beta version for testers. 
#
#                  ** 24.02.2009 v0.9b (Olli Lammi) **
#                  Added support for rrdtool data store. 
#
#                  ** 25.02.2009 v0.9c (Olli Lammi) **
#                  Changed logging. Changed rrdtool support to be
#                  not dependent on external rrdtool-python library.
#                  Changed module imports not to import requirement
#                  modules if a module is not enabled at the configuration. 
#
#                  ** 02.03.2009 v0.9d (Olli Lammi) **
#                  Added verbose logging and daemon configuration directive.
#                  Added pattern support for file data store. Changed the
#                  way oumanSerial accesses the serial port and added 
#                  file locking for the port access.
#
#                  ** 03.03.2009 v0.9e (Olli Lammi) **
#                  Changed the way how empty measured values are handled.
#
#                  ** 06.04.2009 v1.0a (Olli Lammi) **
#                  First release candidate version. Moved configuration objects to the
#                  modules. If no values are available, completely empty data 
#                  is not stored to datastores. Added preliminary config checks.
#                  Added support for DigiTemp. 
#
#                  ** 06.04.2009 v1.0b (Olli Lammi) **
#                  Correction to DigiTemp tool output handling.
#
#                  ** 02.06.2009 v1.0c (Olli Lammi) **
#                  Added Ouman device profile for 202G controller (Geopro GS-pumps)
#
#                  ** 29.04.2010 v1.1a (Olli Lammi) **
#                  Added Thermia serial module.
#
#                  ** 04.09.2010 v1.1b (Olli Lammi) **
#                  Corrections to Thermia functionality (bugs found in beta testing)
#
#                  ** 22.10.2010 v1.1c (Olli Lammi) **
#                  Added SQLite support.
#
#                  ** 16.11.2010 v1.3a (Olli Lammi) **
#                  Added Rego serial module.
#
###########################################################################

# Imports

import sys, os 
import string,re,time  
import threads 
import traceback

import configuration
import oumanSerial
import thermiaSerial
import regoSerial
import oneWire
import log
import store, storeDb, storeFile, storeRrd, storeDbSQLite

###########################################################################

# Constants

VERSION = "v1.3a"

DEFAULT_LOG_INTERVAL = 120
DEFAULT_LOG_INTERVAL_LIMIT = 5


###########################################################################

# Globals

main_terminated = 0

LOG = None


###########################################################################

# Classes

class TaloLoggerThread(threads.Thread, log.Logging):
    def __init__(self, meas, sources, stores):
        threads.Thread.__init__(self)
        log.Logging.__init__(self, 'TaloLoggerThread')

        self.inqueue = []
        self.inqueue_item = []
        self.inq_lock = threads.Lock()
        self.datasourcelist = sources
        self.storelist = stores
        self.measures = meas
        self.terminated = 0
        self.result = {}

    def terminate(self):
        self.terminated = 1

    def isterminated(self):
        return self.terminated

    def received(self, key, data):
        if self.inq_lock.lock_wait():
            if len(self.inqueue_item) > 0:
                if key in self.inqueue_item[0][1]:
                    if key in self.result:
                        self.result[key] = self.result[key] + data
                    else:
                        self.result[key] = data
                    status = 1
                    for k in self.inqueue_item[0][1]:
                        if k not in self.result:
                            status = 0
                    if status:
                        item = self.inqueue_item[0]
                        self.inqueue_item = self.inqueue_item[1:]
                        self.handleResult(item[0], self.result)
                        self.result = {}
                else:
                    self.Log("Received not queried data: " + key)
            else:
                self.Log("Received not queried data: " + key)
            self.inq_lock.free()

    def handleResult(self, type, data):
        if type == 'log':
            self.handleLogResult(data)
        else:
            self.Log("Received unknown item type " + type);
            for k in data.keys():
                self.Log("  " + k + ":" + data[k])
        
    def handleLogResult(self, data):
        self.Debug("Storing logged data with " + repr(len(data)) + " points.")
        # map the result data to keys and values
        result = []
        num = 0
        for item in self.measures:
            temps = data[item[1]]
            if len(temps) > 0:
                num = num + 1
            result.append([item[0], temps])
        if len(self.storelist) > 0 and num > 0:
            for store in self.storelist:
                store.insertData(result)
        else:
            temps = ""
            for item in result:
                temps = temps + "\n    " + item[0] + ": " + item[1]
            self.Log(temps)

    def queue(self, type, val):
        if self.inq_lock.lock_wait():
            self.inqueue_item.append([type, val])
            for v in val:
                self.inqueue.append(v)
            self.inq_lock.free()

    def run(self):
        while not self.terminated:
            try:
                # check if there is something to do
                if self.inq_lock.lock_wait():
                    temp = ""
                    if len(self.inqueue) > 0:
                        temp = self.inqueue[0]
                        self.inqueue = self.inqueue[1:]
                    self.inq_lock.free()
                    if len(temp) > 0:
                        retries = 3
                        data = ""
                        while len(data) <= 0 and retries > 0:
                          temp2 = string.split(temp, '.', 1)
                          fnd = 0
                          for s in self.datasourcelist:
                              if s.getServiceName() == temp2[0]:
                                  fnd = 1
                                  data = s.runDataSourceQueryCommand(temp2[1])
                          if not fnd:
                              self.Log("Unknown command source " + temp)
                              break
                          if len(data) <= 0:
                              retries = retries - 1
                              if retries > 0:
                                  self.Log("Received none or unknown data for command " + temp + ". Retrying...");
                                  time.sleep(0.2)
                              else:
                                  self.Log("Received none or unknown data for command " + temp + ". Giving up...");
                        self.received(temp, data)
                    else:
                        time.sleep(1)
            except Exception, e:
                traceback.print_exc(file=sys.stdout)
                self.Log("Exception: " + e.__str__())
                self.terminated = 1



class DataSource:
    def getServiceName(self): pass

    def runDataSourceQueryCommand(self, cmd): pass


class OumanSerialConf(oumanSerial.OumanSerial, configuration.Configurable, DataSource):
    def __init__(self, conf):
        self.handleConfiguration(conf)
        oumanSerial.OumanSerial.__init__(self, self.port, self.device)

    def getServiceName(self):
        return "OUMAN"

    def runDataSourceQueryCommand(self, cmd):
        return self.runQueryCommand(cmd)

    def handleConfiguration(self, conf):
        self.port = conf.getValue('OUMAN_SERIAL_PORT', '')
        self.device = conf.getValue('OUMAN_DEVICE', '')
        if len(self.port) <= 0:
            return (0, 'Missing Ouman serial port path.')
        if len(self.device) <= 0:
            return (0, 'Missing Ouman device name.')
        return (1, '')
        
    def initConfiguration(self):
        temps = oumanSerial.getOumanDataPart(self.runQueryId(1))
        if len(temps) <= 0:
            return (0, 'Error reading device id from Ouman controller.')
        else:
            if len(temps) > 6:
                temps = temps[6:]
            self.Log('Received Ouman device ID: ' + temps)
        return (1, '')

    @staticmethod
    def getAllowedConfigurationKeys():
        return (['OUMAN_DEVICE', 'OUMAN_SERIAL_PORT'], [])


class ThermiaSerialConf(thermiaSerial.ThermiaSerial, configuration.Configurable, DataSource):
    def __init__(self, conf):
        self.handleConfiguration(conf)
        thermiaSerial.ThermiaSerial.__init__(self, self.port, self.device)

    def getServiceName(self):
        return "THERMIA"

    def runDataSourceQueryCommand(self, cmd):
        return self.runQueryCommand(cmd)

    def handleConfiguration(self, conf):
        self.port = conf.getValue('THERMIA_SERIAL_PORT', '')
        self.device = conf.getValue('THERMIA_DEVICE', '')
        if len(self.port) <= 0:
            return (0, 'Missing Thermia serial port path.')
        if len(self.device) <= 0:
            return (0, 'Missing Thermia device name.')
        return (1, '')
        
    def initConfiguration(self):
        temps = self.getVersion()
        if len(temps) <= 0:
            return (0, 'Error reading version from Thermia controller.')
        else:
            self.Log('Received Thermia device version: ' + temps)
        return (1, '')

    @staticmethod
    def getAllowedConfigurationKeys():
        return (['THERMIA_DEVICE', 'THERMIA_SERIAL_PORT'], [])


class RegoSerialConf(regoSerial.RegoSerial, configuration.Configurable, DataSource):
    def __init__(self, conf):
        self.handleConfiguration(conf)
        regoSerial.RegoSerial.__init__(self, self.port, self.device)

    def getServiceName(self):
        return "REGO"

    def runDataSourceQueryCommand(self, cmd):
        return self.runQueryCommand(cmd)

    def handleConfiguration(self, conf):
        self.port = conf.getValue('REGO_SERIAL_PORT', '')
        self.device = conf.getValue('REGO_DEVICE', '')
        if len(self.port) <= 0:
            return (0, 'Missing Rego serial port path.')
        if len(self.device) <= 0:
            return (0, 'Missing Rego device name.')
        return (1, '')
        
    def initConfiguration(self):
        temps = self.getVersion()
        if len(temps) <= 0:
            return (0, 'Error reading version from Rego controller.')
        else:
            self.Log('Received Rego device version: ' + temps)
        return (1, '')

    @staticmethod
    def getAllowedConfigurationKeys():
        return (['REGO_DEVICE', 'REGO_SERIAL_PORT'], [])


class OneWireFSConf(oneWire.OneWireFS, configuration.Configurable, DataSource):
    def __init__(self, conf):
        oneWire.OneWireFS.__init__(self)
        self.handleConfiguration(conf)

    def getServiceName(self):
        return "OWFS"

    def handleConfiguration(self, conf):
        self.setDevices( self.handleOWFSConf(conf) )
        return (1, '')
        
    def initConfiguration(self):
        return (1, '')

    def handleOWFSConf(self, conf):
        devmap = {}
        for item in conf.getValue('OWFS', []):
            temps = string.split(item, ':', 2)
            if len(temps) < 3:
                self.log("ERROR: Invalid OWFS configuration: " + item)
            else:
                try:
                    key = string.strip(temps[0])
                    type = int(string.strip(temps[1]))
                    value = string.strip(temps[2])
                except:
                    type = 0
                if len(key) <= 0 or len(value) <= 0 or type <= 0:
                    self.log("ERROR: Invalid OWFS configuration: " + item)
                else:
                    devmap[key] = [type, value]
        return devmap

    def runDataSourceQueryCommand(self, cmd):
        return self.runQueryCommand(cmd)

    @staticmethod
    def getAllowedConfigurationKeys():
        return ([], ['OWFS'])


class OneWireDigiTempConf(oneWire.OneWireDigiTemp, configuration.Configurable, DataSource):
    def __init__(self, conf):
        oneWire.OneWireDigiTemp.__init__(self, '', '', '')
        self.handleConfiguration(conf)

    def getServiceName(self):
        return "DIGITEMP"

    def handleConfiguration(self, conf): 
        self.DT_BINARY = conf.getValue('DIGITEMP_LOCATION', '')
        self.DT_RC = conf.getValue('DIGITEMP_CONFFILE', '')
        self.DT_PORT = conf.getValue('DIGITEMP_PORT', '')
        return (1, '')
        
    def initConfiguration(self):
        # check the existence of the digitemp executable
        if not os.access(self.DT_BINARY, os.X_OK):
            return (0, 'OneWireDigiTemp: Unable to find or execute DigiTemp executable: ' + self.DT_BINARY)
        if not os.access(self.DT_RC, os.R_OK):
            return (0, 'OneWireDigiTemp: Unable to access DigiTemp configuration file: ' + self.DT_RC)
        if len(self.DT_PORT) <= 0:
            return (0, 'OneWireDigiTemp: Unspecified DigiTemp port.')
        return (1, '')

    def runDataSourceQueryCommand(self, cmd):
        return self.runQueryCommand(cmd)

    @staticmethod
    def getAllowedConfigurationKeys():
        return (['DIGITEMP_LOCATION', 'DIGITEMP_PORT', 'DIGITEMP_CONFFILE'], [])


###########################################################################

# Functions

def HUPhandler(signum, frame):
  global main_terminated
  main_terminated = 1
  return

def TERMhandler(signum, frame):
  global main_terminated, LOG
  LOG.log("Terminating due to SIGTERM.")
  main_terminated = 1
  return

####### main ##########################################################

def main():
  global main_terminated, LOG

  CONFIGURATION_FILE = "taloLogger.conf"

  daemon = 0
  logconsole = 0
  verbose = 0

  # handle command line parameters
  i = 1
  while i < len(sys.argv):
      if sys.argv[i] == '--help':
          Help()
          sys.exit(1)
      elif sys.argv[i] == '-l':
          logconsole = 1
      elif sys.argv[i] == '-v':
          verbose = 1
      elif sys.argv[i] == '-d':
          daemon = 1
      elif  sys.argv[i] == '-f' and i+1 < len(sys.argv):
          i = i + 1
          CONFIGURATION_FILE = sys.argv[i]
      else:
          Usage()
          sys.exit(1)

      i = i + 1 

  # load configuration file
  conf = configuration.Configuration()
  
  conf.addConfigurable(log.Logger)
  conf.addConfigurable(OumanSerialConf)
  conf.addConfigurable(ThermiaSerialConf)
  conf.addConfigurable(RegoSerialConf)
  conf.addConfigurable(OneWireFSConf)
  conf.addConfigurable(OneWireDigiTempConf)
  conf.addConfigurable(storeDb.DBStore)
  conf.addConfigurable(storeDbSQLite.DBStoreSQLite)
  conf.addConfigurable(storeFile.FileStore)
  conf.addConfigurable(storeRrd.RrdStore)

  conf.addAllowedKeys(['LOG_INTERVAL', 'LOG_INTERVAL_LIMIT', 'DAEMON_MODE', \
                       'ENABLE_OUMAN', 'ENABLE_THERMIA', 'ENABLE_REGO', \
                       'ENABLE_OWFS', 'ENABLE_DIGITEMP', \
                       'ENABLE_MYSQL_DB_STORE', 'ENABLE_FILE_STORE', 'ENABLE_RRD_STORE', 
                       'ENABLE_SQLITE_DB_STORE'])
  conf.addAllowedListKeys(['MEASURE'])

  (confstat, errmsg) = conf.loadFile(CONFIGURATION_FILE)
  if confstat == 0:
      print "ERROR: Cannot read configuration file: " + CONFIGURATION_FILE
      sys.exit(1)
  elif confstat == -1:
      print "ERROR: Error reading configuration file: " + CONFIGURATION_FILE
      print errmsg
      sys.exit(1)

  if logconsole:
      conf.setValue('CONSOLE_LOGGING', 'true')
  if verbose:
      conf.setValue('VERBOSE_LOGGING', 'true')

  LOG = log.Logger(conf)
  log.Logging.setLogger(LOG)

  sources = []
  oser = None
  if conf.isTrue('ENABLE_OWFS'):
      sources.append( OneWireFSConf(conf) )
  if conf.isTrue('ENABLE_DIGITEMP'):
      sources.append( OneWireDigiTempConf(conf) )
  if conf.isTrue('ENABLE_OUMAN'):
      sources.append( OumanSerialConf(conf) )
  if conf.isTrue('ENABLE_THERMIA'):
      sources.append( ThermiaSerialConf(conf) )
  if conf.isTrue('ENABLE_REGO'):
      sources.append( RegoSerialConf(conf) )

  if len(sources) <= 0:
      LOG.log("ERROR: No data sources configured. Nothing to log. Exiting.")
      sys.exit(1)

  stores = []
  if conf.isTrue('ENABLE_FILE_STORE'):
      stores.append( storeFile.FileStore(conf) )
  if conf.isTrue('ENABLE_MYSQL_DB_STORE'):
      stores.append( storeDb.DBStore(conf) )
  if conf.isTrue('ENABLE_SQLITE_DB_STORE'):
      stores.append( storeDbSQLite.DBStoreSQLite(conf) )
  if conf.isTrue('ENABLE_RRD_STORE'):
      stores.append( storeRrd.RrdStore(conf) )

  if len(stores) <= 0:
      LOG.log("WARNING: No log data stores configured. Data is written to application log.")

  for confmod in sources + stores:
      (stat, msg) = confmod.initConfiguration()
      if stat == 0:
          LOG.log("ERROR: Error initializing modules. Exiting.: " + msg)
          print "ERROR: Error initializing modules. Exiting.: " + msg
          sys.exit(1)
          
  if daemon or conf.isTrue('DAEMON_MODE'):
      pid = os.fork()
      if (pid != 0):
          LOG.log("Starting taloLogger.py (" + VERSION + ") as daemon.")
          sys.exit(0) 
  else:
      LOG.log("Starting taloLogger.py (" + VERSION + ").")

  # try to enable signal handlers, if fails, we are probably using Windows
  try:
      import signal
      signal.signal(signal.SIGHUP, HUPhandler) 
      signal.signal(signal.SIGTERM, TERMhandler)
  except:
      pass

  LOG_INTERVAL = int(conf.getValue('LOG_INTERVAL', DEFAULT_LOG_INTERVAL))
  LOG_INTERVAL_LIMIT = int(conf.getValue('LOG_INTERVAL_LIMIT', DEFAULT_LOG_INTERVAL_LIMIT))

  measurekeys = []
  measures = []
  for item in conf.getValue('MEASURE', []):
      temps = string.split(item, ':', 1)
      if len(temps) < 2:
          LOG.log("ERROR: Invalid MEASURE configuration: " + item)
          sys.exit(1)          
      key = string.strip(temps[0])
      value = string.strip(temps[1])
      if len(key) <= 0 or len(value) <= 0:
          LOG.log("ERROR: Invalid MEASURE configuration: " + item)
          sys.exit(1)                    
      measurekeys.append(value)
      measures.append([key, value])
  if len(measures) <= 0:
      LOG.log("ERROR: No configured measures. Nothing to log. Exiting.")
      sys.exit(1)                    

  loggerthread = TaloLoggerThread(measures, sources, stores)

  if not loggerthread.start():
      LOG.log("ERROR: Cannot start logger thread.")
      sys.exit(1)

  while not main_terminated and not loggerthread.isterminated():
      t = time.time() % LOG_INTERVAL
      if t < LOG_INTERVAL_LIMIT:
          loggerthread.queue('log', measurekeys)
          t = time.time() % LOG_INTERVAL
      time.sleep(LOG_INTERVAL - t)

  loggerthread.terminate() 
  loggerthread.wait_sleeping()

  LOG.log("Terminating.")


def Usage():
    print 'USAGE:  python taloLogger.py --help'
    print '        python taloLogger.py [-d] [-v] [-l] [-f configuration_file]'


def Help():
    temps = string.split(HELP_TEXT, '\n')
    for temp in temps:
        if len(temp) > 0 and temp[0] == '#':
            temp = temp[1:]
        if len(temp) > 0 and temp[0] == ' ':
            temp = temp[1:]
        print temp
 

###########################################################################

if __name__ == "__main__":
  main()

